@@ -1,11 +1,13 @@
+use std::collections::HashMap;
 use std::iter::FromIterator;
 use std::panic::{catch_unwind, AssertUnwindSafe};
+use std::string::String as StdString;
 use std::sync::Arc;
 use std::{error, f32, f64, fmt};
 
 use mlua::{
-    ChunkMode, Error, ExternalError, Function, Lua, Nil, Result, StdLib, String, Table, UserData,
-    Value, Variadic,
+    ChunkMode, Error, ExternalError, Function, Lua, LuaOptions, Nil, Result, StdLib, String, Table,
+    UserData, Value, Variadic,
 };
 
 #[test]
@@ -23,7 +25,7 @@ fn test_safety() -> Result<()> {
     assert!(lua.load(r#"require "debug""#).exec().is_ok());
     drop(lua);
 
-    match Lua::new_with(StdLib::DEBUG) {
+    match Lua::new_with(StdLib::DEBUG, LuaOptions::default()) {
         Err(Error::SafetyError(_)) => {}
         Err(e) => panic!("expected SafetyError, got {:?}", e),
         Ok(_) => panic!("expected SafetyError, got new Lua state"),
@@ -52,7 +54,7 @@ fn test_safety() -> Result<()> {
         Ok(_) => panic!("expected SafetyError, got no error"),
     }
 
-    let bytecode = lua.load("return 1 + 1").into_function()?.dump(true)?;
+    let bytecode = lua.load("return 1 + 1").into_function()?.dump(true);
     match lua.load(&bytecode).exec() {
         Err(Error::SafetyError(msg)) => {
             assert!(msg.contains("binary chunks are disabled in safe mode"))
@@ -63,7 +65,7 @@ fn test_safety() -> Result<()> {
     drop(lua);
 
     // Test safety rules after dynamically loading `package` library
-    let lua = Lua::new_with(StdLib::NONE)?;
+    let lua = Lua::new_with(StdLib::NONE, LuaOptions::default())?;
     assert!(lua.globals().get::<_, Option<Value>>("require")?.is_none());
     lua.load_from_std_lib(StdLib::PACKAGE)?;
     match lua.load(r#"package.loadlib()"#).exec() {
@@ -81,6 +83,7 @@ fn test_safety() -> Result<()> {
 #[test]
 fn test_load() -> Result<()> {
     let lua = Lua::new();
+
     let func = lua.load("return 1+2").into_function()?;
     let result: i32 = func.call(())?;
     assert_eq!(result, 3);
@@ -162,7 +165,7 @@ fn test_load_mode() -> Result<()> {
         Err(e) => panic!("expected SyntaxError, got {:?}", e),
     };
 
-    let bytecode = lua.load("return 1 + 1").into_function()?.dump(true)?;
+    let bytecode = lua.load("return 1 + 1").into_function()?.dump(true);
     assert_eq!(lua.load(&bytecode).eval::<i32>()?, 2);
     assert_eq!(
         lua.load(&bytecode)
@@ -221,6 +224,7 @@ fn test_coercion() -> Result<()> {
         int = 123
         str = "123"
         num = 123.0
+        func = function() end
     "#,
     )
     .exec()?;
@@ -229,6 +233,7 @@ fn test_coercion() -> Result<()> {
     assert_eq!(globals.get::<_, String>("int")?, "123");
     assert_eq!(globals.get::<_, i32>("str")?, 123);
     assert_eq!(globals.get::<_, i32>("num")?, 123);
+    assert!(globals.get::<_, String>("func").is_err());
 
     Ok(())
 }
@@ -377,18 +382,23 @@ fn test_error() -> Result<()> {
 
 #[test]
 fn test_panic() -> Result<()> {
-    fn make_lua() -> Result<Lua> {
-        let lua = Lua::new();
+    fn make_lua(options: LuaOptions) -> Result<Lua> {
+        let lua = Lua::new_with(StdLib::ALL_SAFE, options)?;
         let rust_panic_function =
-            lua.create_function(|_, ()| -> Result<()> { panic!("rust panic") })?;
+            lua.create_function(|_, msg: Option<StdString>| -> Result<()> {
+                if let Some(msg) = msg {
+                    panic!("{}", msg)
+                }
+                panic!("rust panic")
+            })?;
         lua.globals()
             .set("rust_panic_function", rust_panic_function)?;
         Ok(lua)
     }
 
-    // Test triggerting Lua error passing Rust panic (must be resumed)
+    // Test triggering Lua error with sending Rust panic (must be resumed)
     {
-        let lua = make_lua()?;
+        let lua = make_lua(LuaOptions::default())?;
 
         match catch_unwind(AssertUnwindSafe(|| -> Result<()> {
             lua.load(
@@ -414,7 +424,7 @@ fn test_panic() -> Result<()> {
 
     // Test returning Rust panic (must be resumed)
     {
-        let lua = make_lua()?;
+        let lua = make_lua(LuaOptions::default())?;
         match catch_unwind(AssertUnwindSafe(|| -> Result<()> {
             let _catched_panic = lua
                 .load(
@@ -442,9 +452,9 @@ fn test_panic() -> Result<()> {
         }
     }
 
-    // Test representing rust panic as a string
+    // Test representing Rust panic as a string
     match catch_unwind(|| -> Result<()> {
-        let lua = make_lua()?;
+        let lua = make_lua(LuaOptions::default())?;
         lua.load(
             r#"
             local _, err = pcall(rust_panic_function)
@@ -459,6 +469,48 @@ fn test_panic() -> Result<()> {
         Err(_) => panic!("panic was detected"),
     }
 
+    // Test disabling `catch_rust_panics` option / pcall correctness
+    match catch_unwind(|| -> Result<()> {
+        let lua = make_lua(LuaOptions::new().catch_rust_panics(false))?;
+        lua.load(
+            r#"
+            local ok, err = pcall(function(msg) error(msg) end, "hello")
+            assert(not ok and err:find("hello") ~= nil)
+
+            ok, err = pcall(rust_panic_function, "rust panic from lua")
+            -- Nothing to return, panic should be automatically resumed
+        "#,
+        )
+        .exec()
+    }) {
+        Ok(r) => panic!("no panic was detected: {:?}", r),
+        Err(p) => assert!(*p.downcast::<StdString>().unwrap() == "rust panic from lua"),
+    }
+
+    // Test disabling `catch_rust_panics` option / xpcall correctness
+    match catch_unwind(|| -> Result<()> {
+        let lua = make_lua(LuaOptions::new().catch_rust_panics(false))?;
+        lua.load(
+            r#"
+            local msgh_ok = false
+            local msgh = function(err)
+                msgh_ok = err ~= nil and err:find("hello") ~= nil
+                return err
+            end
+            local ok, err = xpcall(function(msg) error(msg) end, msgh, "hello")
+            assert(not ok and err:find("hello") ~= nil)
+            assert(msgh_ok)
+
+            ok, err = xpcall(rust_panic_function, msgh, "rust panic from lua")
+            -- Nothing to return, panic should be automatically resumed
+        "#,
+        )
+        .exec()
+    }) {
+        Ok(r) => panic!("no panic was detected: {:?}", r),
+        Err(p) => assert!(*p.downcast::<StdString>().unwrap() == "rust panic from lua"),
+    }
+
     Ok(())
 }
 
@@ -561,7 +613,7 @@ fn test_pcall_xpcall() -> Result<()> {
     assert!(lua.load("xpcall()").exec().is_err());
     assert!(lua.load("xpcall(function() end)").exec().is_err());
 
-    // Lua 5.3/5.2 / LuaJIT compatible version of xpcall
+    // Lua >= 5.2 compatible version of xpcall for 5.1
     #[cfg(feature = "lua51")]
     lua.load(
         r#"
@@ -747,6 +799,23 @@ fn test_drop_registry_value() -> Result<()> {
     Ok(())
 }
 
+#[test]
+fn test_lua_registry_hash() -> Result<()> {
+    let lua = Lua::new();
+
+    let r1 = Arc::new(lua.create_registry_value("value1")?);
+    let r2 = Arc::new(lua.create_registry_value("value2")?);
+
+    let mut map = HashMap::new();
+    map.insert(r1.clone(), "value1");
+    map.insert(r2.clone(), "value2");
+
+    assert_eq!(map[&r1], "value1");
+    assert_eq!(map[&r2], "value2");
+
+    Ok(())
+}
+
 #[test]
 fn test_lua_registry_ownership() -> Result<()> {
     let lua1 = Lua::new();
@@ -778,7 +847,26 @@ fn test_mismatched_registry_key() -> Result<()> {
 }
 
 #[test]
-fn too_many_returns() -> Result<()> {
+fn test_recursion() -> Result<()> {
+    let lua = Lua::new();
+
+    let f = lua.create_function(move |lua, i: i32| {
+        if i < 64 {
+            lua.globals()
+                .get::<_, Function>("f")?
+                .call::<_, ()>(i + 1)?;
+        }
+        Ok(())
+    })?;
+
+    lua.globals().set("f", f.clone())?;
+    f.call::<_, ()>(1)?;
+
+    Ok(())
+}
+
+#[test]
+fn test_too_many_returns() -> Result<()> {
     let lua = Lua::new();
     let f = lua.create_function(|_, ()| Ok(Variadic::from_iter(1..1000000)))?;
     assert!(f.call::<_, Vec<u32>>(()).is_err());
@@ -786,7 +874,7 @@ fn too_many_returns() -> Result<()> {
 }
 
 #[test]
-fn too_many_arguments() -> Result<()> {
+fn test_too_many_arguments() -> Result<()> {
     let lua = Lua::new();
     lua.load("function test(...) end").exec()?;
     let args = Variadic::from_iter(1..1000000);
@@ -800,23 +888,20 @@ fn too_many_arguments() -> Result<()> {
 
 #[test]
 #[cfg(not(feature = "luajit"))]
-fn too_many_recursions() -> Result<()> {
+fn test_too_many_recursions() -> Result<()> {
     let lua = Lua::new();
+
     let f = lua
         .create_function(move |lua, ()| lua.globals().get::<_, Function>("f")?.call::<_, ()>(()))?;
-    lua.globals().set("f", f)?;
 
-    assert!(lua
-        .globals()
-        .get::<_, Function>("f")?
-        .call::<_, ()>(())
-        .is_err());
+    lua.globals().set("f", f.clone())?;
+    assert!(f.call::<_, ()>(()).is_err());
 
     Ok(())
 }
 
 #[test]
-fn too_many_binds() -> Result<()> {
+fn test_too_many_binds() -> Result<()> {
     let lua = Lua::new();
     let globals = lua.globals();
     lua.load(
@@ -837,7 +922,25 @@ fn too_many_binds() -> Result<()> {
 }
 
 #[test]
-fn large_args() -> Result<()> {
+fn test_ref_stack_exhaustion() {
+    match catch_unwind(AssertUnwindSafe(|| -> Result<()> {
+        let lua = Lua::new();
+        let mut vals = Vec::new();
+        for _ in 0..1000000 {
+            vals.push(lua.create_table()?);
+        }
+        Ok(())
+    })) {
+        Ok(_) => panic!("no panic was detected"),
+        Err(p) => assert!(p
+            .downcast::<StdString>()
+            .unwrap()
+            .starts_with("cannot create a Lua reference, out of auxiliary stack space")),
+    }
+}
+
+#[test]
+fn test_large_args() -> Result<()> {
     let lua = Lua::new();
     let globals = lua.globals();
 
@@ -872,7 +975,7 @@ fn large_args() -> Result<()> {
 }
 
 #[test]
-fn large_args_ref() -> Result<()> {
+fn test_large_args_ref() -> Result<()> {
     let lua = Lua::new();
 
     let f = lua.create_function(|_, args: Variadic<String>| {
@@ -888,7 +991,7 @@ fn large_args_ref() -> Result<()> {
 }
 
 #[test]
-fn chunk_env() -> Result<()> {
+fn test_chunk_env() -> Result<()> {
     let lua = Lua::new();
 
     let assert: Function = lua.globals().get("assert")?;
@@ -930,7 +1033,7 @@ fn chunk_env() -> Result<()> {
 }
 
 #[test]
-fn context_thread() -> Result<()> {
+fn test_context_thread() -> Result<()> {
     let lua = Lua::new();
 
     let f = lua
@@ -953,7 +1056,7 @@ fn context_thread() -> Result<()> {
 
 #[test]
 #[cfg(any(feature = "lua51", feature = "luajit"))]
-fn context_thread_51() -> Result<()> {
+fn test_context_thread_51() -> Result<()> {
     let lua = Lua::new();
 
     let thread = lua.create_thread(
